# This file is part of lsst-resources.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# Use of this source code is governed by a 3-clause BSD-style
# license that can be found in the LICENSE file.

import contextlib
import os
import pathlib
import unittest
import unittest.mock
import urllib.parse

from lsst.resources import ResourcePath, ResourcePathExpression
from lsst.resources.tests import GenericReadWriteTestCase, GenericTestCase

TESTDIR = os.path.abspath(os.path.dirname(__file__))


class SimpleTestCase(unittest.TestCase):
    """Basic tests for file URIs."""

    def test_instance(self):
        for example in (
            "xxx",
            ResourcePath("xxx"),
            pathlib.Path("xxx"),
            urllib.parse.urlparse("file:///xxx"),
        ):
            self.assertIsInstance(example, ResourcePathExpression)

        for example in ({1, 2, 3}, 42, self):
            self.assertNotIsInstance(example, ResourcePathExpression)


class FileTestCase(GenericTestCase, unittest.TestCase):
    """File-specific generic test cases."""

    scheme = "file"
    netloc = "localhost"

    def test_env_var(self):
        """Test that environment variables are expanded."""
        with unittest.mock.patch.dict(os.environ, {"MY_TEST_DIR": "/a/b/c"}):
            uri = ResourcePath("${MY_TEST_DIR}/d.txt")
        self.assertEqual(uri.path, "/a/b/c/d.txt")
        self.assertEqual(uri.scheme, "file")

        # This will not expand
        uri = ResourcePath("${MY_TEST_DIR}/d.txt", forceAbsolute=False)
        self.assertEqual(uri.path, "${MY_TEST_DIR}/d.txt")
        self.assertFalse(uri.scheme)

    def test_ospath(self):
        """File URIs have ospath property."""
        file = ResourcePath(self._make_uri("a/test.txt"))
        self.assertEqual(file.ospath, "/a/test.txt")
        self.assertEqual(file.ospath, file.path)

        # A Schemeless URI can take unquoted files but will be quoted
        # when it becomes a file URI.
        something = "/a#/???.txt"
        file = ResourcePath(something, forceAbsolute=True)
        self.assertEqual(file.scheme, "file")
        self.assertEqual(file.ospath, something, "From URI: {file}")
        self.assertNotIn("???", file.path)

    def test_path_lib(self):
        """File URIs can be created from pathlib."""
        file = ResourcePath(self._make_uri("a/test.txt"))

        path_file = pathlib.Path(file.ospath)
        from_path = ResourcePath(path_file)
        self.assertEqual(from_path.ospath, file.ospath)

    def test_schemeless_root(self):
        root = ResourcePath(self._make_uri("/root"))
        via_root = ResourcePath("b.txt", root=root)
        self.assertEqual(via_root.ospath, "/root/b.txt")


TEST_UMASK = 0o0333


class FileReadWriteTestCase(GenericReadWriteTestCase, unittest.TestCase):
    """File tests involving reading and writing of data."""

    scheme = "file"
    netloc = "localhost"
    testdir = TESTDIR
    transfer_modes = ("move", "copy", "link", "hardlink", "symlink", "relsymlink")

    def test_transfer_identical(self):
        """Test overwrite of identical files.

        Only relevant for local files.
        """
        dir1 = self.tmpdir.join("dir1", forceDirectory=True)
        dir1.mkdir()
        self.assertTrue(dir1.exists())
        dir2 = self.tmpdir.join("dir2", forceDirectory=True)
        # A symlink can't include a trailing slash.
        dir2_ospath = dir2.ospath
        if dir2_ospath.endswith("/"):
            dir2_ospath = dir2_ospath[:-1]
        os.symlink(dir1.ospath, dir2_ospath)

        # Write a test file.
        src_file = dir1.join("test.txt")
        content = "0123456"
        src_file.write(content.encode())

        # Construct URI to destination that should be identical.
        dest_file = dir2.join("test.txt")
        self.assertTrue(dest_file.exists())
        self.assertNotEqual(src_file, dest_file)

        # Transfer it over itself.
        dest_file.transfer_from(src_file, transfer="symlink", overwrite=True)
        new_content = dest_file.read().decode()
        self.assertEqual(content, new_content)

    def test_local_temporary(self):
        """Create temporary local file if no prefix specified."""
        with ResourcePath.temporary_uri(suffix=".json") as tmp:
            self.assertEqual(tmp.getExtension(), ".json", f"uri: {tmp}")
            self.assertTrue(tmp.isabs(), f"uri: {tmp}")
            self.assertFalse(tmp.exists(), f"uri: {tmp}")
            tmp.write(b"abcd")
            self.assertTrue(tmp.exists(), f"uri: {tmp}")
            self.assertTrue(tmp.isTemporary)
            self.assertTrue(tmp.isLocal)

            # If we now ask for a local form of this temporary file
            # it should still be temporary and it should not be deleted
            # on exit.
            with tmp.as_local() as loc:
                self.assertEqual(tmp, loc)
                self.assertTrue(loc.isTemporary)
            self.assertTrue(tmp.exists())
        self.assertFalse(tmp.exists(), f"uri: {tmp}")

        with ResourcePath.temporary_uri(suffix=".yaml", delete=False) as tmp:
            tmp.write(b"1234")
            self.assertTrue(tmp.exists(), f"uri: {tmp}")
        # If the file doesn't exist there is nothing to clean up so a failure
        # here is not a problem.
        self.assertTrue(tmp.exists(), f"uri: {tmp} should still exist")

        # If removal does not work it's worth reporting that as an error.
        tmp.remove()

    def test_transfers_from_local(self):
        """Extra tests for local transfers."""
        target = self.tmpdir.join("a/target.txt")
        with ResourcePath.temporary_uri() as tmp:
            tmp.write(b"")
            self.assertTrue(tmp.isTemporary)

            # Symlink transfers for temporary resources should
            # trigger a debug message.
            for transfer in ("symlink", "relsymlink"):
                with self.assertLogs("lsst.resources", level="DEBUG") as cm:
                    target.transfer_from(tmp, transfer)
                target.remove()
                self.assertIn("Using a symlink for a temporary", "".join(cm.output))

            # Force the target directory to be created.
            target.transfer_from(tmp, "move")
            self.assertFalse(tmp.exists())

            # Temporary file now gone so transfer should not work.
            with self.assertRaises(FileNotFoundError):
                target.transfer_from(tmp, "move", overwrite=True)

    def test_write_with_restrictive_umask(self):
        self._test_file_with_restrictive_umask(lambda target: target.write(b"123"))

    def test_transfer_from_with_restrictive_umask(self):
        def cb(target):
            with ResourcePath.temporary_uri() as tmp:
                tmp.write(b"")
                target.transfer_from(tmp, "copy")

        self._test_file_with_restrictive_umask(cb)

    def test_mkdir_with_restrictive_umask(self):
        self._test_with_restrictive_umask(lambda target: target.mkdir())

    def test_temporary_uri_with_restrictive_umask(self):
        with _override_umask(TEST_UMASK):
            with ResourcePath.temporary_uri() as tmp:
                tmp.write(b"")
                self.assertTrue(tmp.exists())

    def _test_file_with_restrictive_umask(self, callback):
        def inner_cb(target):
            callback(target)

            # Make sure the umask was respected for the file itself
            file_mode = os.stat(target.ospath).st_mode
            self.assertEqual(file_mode & TEST_UMASK, 0)

        self._test_with_restrictive_umask(inner_cb)

    def _test_with_restrictive_umask(self, callback):
        """Make sure that parent directories for a file can be created even if
        the user has set a process umask that restricts the write and traverse
        bits.
        """
        with _override_umask(TEST_UMASK):
            target = self.tmpdir.join("a/b/target.txt")
            callback(target)
            self.assertTrue(target.exists())

            dir_b_path = os.path.dirname(target.ospath)
            dir_a_path = os.path.dirname(dir_b_path)
            for dir in [dir_a_path, dir_b_path]:
                # Make sure we only added the minimum permissions needed for it
                # to work (owner-write and owner-traverse)
                mode = os.stat(dir).st_mode
                self.assertEqual(mode & TEST_UMASK, 0o0300, f"Permissions incorrect for {dir}: {mode:o}")

    @unittest.mock.patch("lsst.resources._resourcePath._POOL_EXECUTOR_CLASS", None)
    @unittest.mock.patch.dict(os.environ, {"LSST_RESOURCES_EXECUTOR": "process"})
    def test_mexists_process(self) -> None:
        """Test mexists with override executor pool.

        Force test with process pool.
        """
        super().test_mexists()

    @unittest.mock.patch("lsst.resources._resourcePath._POOL_EXECUTOR_CLASS", None)
    @unittest.mock.patch.dict(os.environ, {"LSST_RESOURCES_EXECUTOR": "process"})
    def test_mtransfer_process(self) -> None:
        """Test transfer with override executor pool.

        Force test with process pool.
        """
        super().test_mtransfer()


@contextlib.contextmanager
def _override_umask(temp_umask):
    old = os.umask(temp_umask)
    try:
        yield
    finally:
        os.umask(old)


if __name__ == "__main__":
    unittest.main()



# This file is part of lsst-resources.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# Use of this source code is governed by a 3-clause BSD-style
# license that can be found in the LICENSE file.

from __future__ import annotations

__all__ = ("DavReadResourceHandle",)

import io
import logging
from collections.abc import Callable, Iterable
from typing import TYPE_CHECKING, AnyStr

from ..davutils import DavFileMetadata
from ._baseResourceHandle import BaseResourceHandle, CloseStatus

if TYPE_CHECKING:
    from ..dav import DavResourcePath


class DavReadResourceHandle(BaseResourceHandle[bytes]):
    """WebDAV-based specialization of `.BaseResourceHandle`.

    Parameters
    ----------
    mode : `str`
        Handle modes as described in the python `io` module.
    log : `~logging.Logger`
        Logger to used when writing messages.
    uri : `lsst.resources.dav.DavResourcePath`
        URI of remote resource.
    newline : `str` or `None`, optional
        When doing multiline operations, break the stream on given character.
        Defaults to newline. If a file is opened in binary mode, this argument
        is not used, as binary files will only split lines on the binary
        newline representation.
    """

    def __init__(
        self,
        mode: str,
        log: logging.Logger,
        uri: DavResourcePath,
        stat: DavFileMetadata,
        *,
        newline: AnyStr | None = None,
    ) -> None:
        super().__init__(mode, log, uri, newline=newline)
        self._uri: DavResourcePath = uri
        self._stat: DavFileMetadata = stat
        self._current_position = 0
        self._cache: io.BytesIO | None = None
        self._buffer: io.BytesIO | None = None
        self._closed = CloseStatus.OPEN

    def close(self) -> None:
        self._closed = CloseStatus.CLOSED
        self._cache = None

    @property
    def closed(self) -> bool:
        return self._closed == CloseStatus.CLOSED

    def fileno(self) -> int:
        raise io.UnsupportedOperation("DavReadResourceHandle does not have a file number")

    def flush(self) -> None:
        modes = set(self._mode)
        if {"w", "x", "a", "+"} & modes:
            raise io.UnsupportedOperation("DavReadResourceHandles are read only")

    @property
    def isatty(self) -> bool | Callable[[], bool]:
        return False

    def readable(self) -> bool:
        return True

    def readline(self, size: int = -1) -> bytes:
        raise io.UnsupportedOperation("DavReadResourceHandles Do not support line by line reading")

    def readlines(self, size: int = -1) -> Iterable[bytes]:
        raise io.UnsupportedOperation("DavReadResourceHandles Do not support line by line reading")

    def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
        match whence:
            case io.SEEK_SET:
                if offset < 0:
                    raise ValueError(f"negative seek value {offset}")
                self._current_position = offset
            case io.SEEK_CUR:
                self._current_position += offset
            case io.SEEK_END:
                self._current_position = self._stat.size + offset
            case _:
                raise ValueError(f"unexpected value {whence} for whence in seek()")

        if self._current_position < 0:
            self._current_position = 0

        return self._current_position

    def seekable(self) -> bool:
        return True

    def tell(self) -> int:
        return self._current_position

    def truncate(self, size: int | None = None) -> int:
        raise io.UnsupportedOperation("DavReadResourceHandles Do not support truncation")

    def writable(self) -> bool:
        return False

    def write(self, b: bytes, /) -> int:
        raise io.UnsupportedOperation("DavReadResourceHandles are read only")

    def writelines(self, b: Iterable[bytes], /) -> None:
        raise io.UnsupportedOperation("DavReadResourceHandles are read only")

    @property
    def _eof(self) -> bool:
        return self._current_position >= self._stat.size

    def _download_to_cache(self) -> io.BytesIO:
        """Download the entire content of the remote resource to an internal
        memory buffer.
        """
        if self._cache is None:
            self._cache = io.BytesIO()
            self._cache.write(self._uri.read())

        return self._cache

    def read(self, size: int = -1) -> bytes:
        if self._eof or size == 0:
            return b""

        # If this file's size is small than the buffer size configured for
        # this URI's client, download the entire file in one request and cache
        # its content. This avoids multiple roundtrips to the server
        # for retrieving small chunks.
        if self._stat.size <= self._uri._client._config.buffer_size:
            self._download_to_cache()

        # If we are asked to read the whole file content, cache the entire
        # file content and return a copy-on-write memory view of our internal
        # cache.
        if self._current_position == 0 and size == -1:
            cache = self._download_to_cache()
            self._current_position = self._stat.size
            return cache.getvalue()

        # This is a partial read. If we have already cached the whole file
        # content use the cache to build the return value.
        if self._cache is not None:
            start = self._current_position
            end = self._current_position = self._stat.size if size < 0 else start + size
            return self._cache.getvalue()[start:end]

        # We need to make a partial read from the server. Reuse our internal
        # I/O buffer to reduce memory allocations.
        if self._buffer is None:
            self._buffer = io.BytesIO()

        start = self._current_position
        end = self._stat.size if size < 0 else min(start + size, self._stat.size)
        self._buffer.seek(0)
        self._buffer.write(self._uri.read_range(start=start, end=end - 1))
        count = self._buffer.tell()
        self._current_position += count
        return self._buffer.getvalue()[0:count]

    def readinto(self, output: bytearray) -> int:
        """Read up to `len(output)` bytes into `output` and return the number
        of bytes read.
        """
        if self._eof or len(output) == 0:
            return 0

        data = self.read(len(output))
        output[:] = data
        return len(data)



# This file is part of lsst-resources.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# Use of this source code is governed by a 3-clause BSD-style
# license that can be found in the LICENSE file.

from __future__ import annotations

__all__ = ("DavResourcePath",)

import contextlib
import datetime
import functools
import io
import logging
import os
import re
import threading
import urllib
from collections.abc import Iterator
from typing import TYPE_CHECKING, Any, BinaryIO, cast

try:
    import fsspec
    from fsspec.spec import AbstractFileSystem
except ImportError:
    fsspec = None
    AbstractFileSystem = type

from ._resourceHandles import ResourceHandleProtocol
from ._resourceHandles._davResourceHandle import DavReadResourceHandle
from ._resourcePath import ResourcePath, ResourcePathExpression
from .davutils import (
    DavClient,
    DavClientPool,
    DavConfigPool,
    DavFileMetadata,
    normalize_path,
    normalize_url,
)
from .utils import get_tempdir

if TYPE_CHECKING:
    from .utils import TransactionProtocol


log = logging.getLogger(__name__)


@functools.lru_cache
def _calc_tmpdir_buffer_size(tmpdir: str) -> int:
    """Compute the block size to use for writing files in `tmpdir` as
    256 blocks of typical size (i.e. 4096 bytes) or 10 times the file system
    block size, whichever is higher.

    This is a reasonable compromise between using memory for buffering and
    the number of system calls issued to read from or write to temporary
    files.
    """
    fsstats = os.statvfs(tmpdir)
    return max(10 * fsstats.f_bsize, 256 * 4096)


class DavResourcePathConfig:
    """Configuration class to encapsulate the configurable items used by
    all instances of class `DavResourcePath`.

    Instantiating this class creates a thread-safe singleton.
    """

    _instance = None
    _lock = threading.Lock()

    def __new__(cls) -> DavResourcePathConfig:
        if cls._instance is None:
            with cls._lock:
                if cls._instance is None:
                    cls._instance = super().__new__(cls)

        return cls._instance

    def __init__(self) -> None:
        # Path to the local temporary directory all instances of
        # `DavResourcePath`must use and its associated buffer size (in bytes).
        self._tmpdir_buffersize: tuple[str, int] | None = None

    @property
    def tmpdir_buffersize(self) -> tuple[str, int]:
        """Return the path to a temporary directory and the preferred buffer
        size to use when reading/writing files from/to that directory.
        """
        if self._tmpdir_buffersize is not None:
            return self._tmpdir_buffersize

        # Retrieve and cache the path and the blocksize for the temporary
        # directory if no other thread has done that in the meantime.
        with DavResourcePathConfig._lock:
            if self._tmpdir_buffersize is None:
                tmpdir = get_tempdir()
                bufsize = _calc_tmpdir_buffer_size(tmpdir)
                self._tmpdir_buffersize = (tmpdir, bufsize)

        return self._tmpdir_buffersize

    def _destroy(self) -> None:
        """Destroy this class singleton instance.

        Helper method to be used in tests to reset global configuration.
        """
        with DavResourcePathConfig._lock:
            DavResourcePathConfig._instance = None


class DavGlobals:
    """Helper container to encapsulate all the gloal objects needed by this
    module.
    """

    def __init__(self) -> None:
        # Client pool used by all DavResourcePath instances.
        # Use Any as type annotation to keep mypy happy.
        self._client_pool: Any = None

        # Configuration used by all DavResourcePath instances.
        self._config: Any = None

        # (Re)Initialize the objects above.
        self._reset()

    def _reset(self) -> None:
        """
        Initialize all the globals.

        This method is a helper for reinitializing globals in tests.
        """
        # Initialize the singleton instance of the webdav endpoint
        # configuration pool.
        config_pool: DavConfigPool = DavConfigPool("LSST_RESOURCES_WEBDAV_CONFIG")

        # Initialize the singleton instance of the webdav client pool. This is
        # a thread-safe singleton shared by all instances of DavResourcePath.
        if self._client_pool is not None:
            self._client_pool._destroy()

        self._client_pool = DavClientPool(config_pool)

        # Initialize the singleton instance of the configuration shared
        # all DavResourcePath objects.
        if self._config is not None:
            self._config._destroy()

        self._config = DavResourcePathConfig()

    def client_pool(self) -> DavClientPool:
        """Return the pool of reusable webDAV clients."""
        return self._client_pool

    def config(self) -> DavResourcePathConfig:
        """Return the configuration settings for all `DavResourcePath`
        objects.
        """
        return self._config


# Convenience object to encapsulate all global objects needed by this module.
dav_globals: DavGlobals = DavGlobals()


class DavResourcePath(ResourcePath):
    """WebDAV resource."""

    def __init__(
        self,
        uri: ResourcePathExpression,
        root: str | ResourcePath | None = None,
        forceAbsolute: bool = True,
        forceDirectory: bool | None = None,
        isTemporary: bool | None = None,
    ) -> None:
        # Build the internal URL we use to talk to the server, which
        # uses "http" or "https" as scheme instead of "dav" or "davs".
        self._internal_url: str = normalize_url(self.geturl())

        # WebDAV client this path must use to interact with the server.
        self._dav_client: DavClient | None = None

        # Retrieve the configuration shared by all instances of this class.
        self._config: DavResourcePathConfig = dav_globals.config()

        # Cached attributes of this file
        self._cached_metadata: DavFileMetadata | None = None

    @classmethod
    def _fixupPathUri(
        cls,
        parsed: urllib.parse.ParseResult,
        root: ResourcePath | None = None,
        forceAbsolute: bool = False,
        forceDirectory: bool | None = None,
    ) -> tuple[urllib.parse.ParseResult, bool | None]:
        """Correct any issues with the supplied URI.

        This function ensures that the path of the URI is normalized.
        """
        # Call the superclass' _fixupPathUri.
        parsed, dirLike = super()._fixupPathUri(parsed, forceDirectory=forceDirectory)

        # Clean the URL's path and ensure dir-like paths end by "/".
        path = normalize_path(parsed.path)
        if dirLike and path != "/":
            path += "/"

        return parsed._replace(path=path), dirLike

    @property
    def _client(self) -> DavClient:
        """Return the webDAV client for this resource."""
        # If we already have a client, use it.
        if self._dav_client is not None:
            return self._dav_client

        # Retrieve the client this resource must use to interact with the
        # server from the global client pool.
        self._dav_client = dav_globals.client_pool().get_client_for_url(self._internal_url)
        return self._dav_client

    def _stat(self, refresh: bool = False) -> DavFileMetadata:
        """Retrieve metadata about this resource.

        We cache this resource's metadata to avoid expensive roundtrips to
        the server for each call.

        Parameters
        ----------
        refresh : `bool`, optional
            If True, metadata is retrieved again from the server even if it
            is already cached.

        Notes
        -----
        Cached metadata is explicitly invalidated when this resource is
        modified, for instance as a result of calling write(), transfer_from()
        remove(), etc.
        """
        # Caching metadata is a compromise because each roundtrip is
        # relatively expensive and is fragile if this same resource is
        # modified by a different thread or by a different process.
        if refresh or self._cached_metadata is None:
            self._cached_metadata = self._client.stat(self._internal_url)

        return self._cached_metadata

    def _invalidate_metatada_cache(self) -> None:
        """Invalidate cached metadata for this resource.

        This method is intended to be explicitly invoked when a method
        modifies the content of this resource (e.g. write, remove,
        transfer_from).
        """
        self._cached_metadata = None

    def mkdir(self) -> None:
        """Create the directory resource if it does not already exist."""
        if not self.isdir():
            raise NotADirectoryError(f"Can not create a directory for file-like URI {self}")

        stat = self._stat()
        if stat.is_dir:
            return

        if stat.is_file:
            # A file exists at this path.
            raise NotADirectoryError(
                f"Can not create a directory for {self} because a file already exists at that URL"
            )

        # Target directory does not exist. Create it and its ancestors as
        # needed. We need to test if parent URL is different from self URL,
        # otherwise we could be stuck in a recursive loop
        # where self == parent.
        if self.geturl() != self.parent().geturl():
            self.parent().mkdir()

        self._client.mkcol(self._internal_url)
        self._invalidate_metatada_cache()

    def exists(self) -> bool:
        """Check that this resource exists."""
        # Force checking for existence against the server for all the
        # external calls to this method.
        return self._stat(refresh=True).exists

    def size(self) -> int:
        """Return the size of the remote resource in bytes."""
        if self.isdir():
            return 0

        stat = self._stat()
        if not stat.exists:
            raise FileNotFoundError(f"No file or directory found at {self}")

        return stat.size

    def info(self) -> dict[str, Any]:
        """Return metadata details about this resource."""
        return self._client.info(self._internal_url, name=str(self))

    def read(self, size: int = -1) -> bytes:
        """Open the resource and return the contents in bytes.

        Parameters
        ----------
        size : `int`, optional
            The number of bytes to read. Negative or omitted indicates that
            all data should be read.
        """
        # A GET request on a dCache directory returns the contents of the
        # directory in HTML, to be visualized with a browser. This means
        # that we need to check first that this resource is not a directory.
        #
        # Since isdir() only checks that the URL of the resource ends in "/"
        # without actually asking the server, this check is not robust.
        # However, it is a reasonable compromise since it prevents doing
        # an additional roundtrip to the server to retrieve this resource's
        # metadata.
        if self.isdir():
            raise ValueError(f"method read() is not implemented for directory {self}")

        stat = self._stat()
        if stat.is_dir:
            raise ValueError(f"method read() is not implemented for directory {self}")
        elif not stat.exists:
            raise FileNotFoundError(f"no file found at {self}")
        elif stat.size == 0:
            # This is an empty file.
            return b""

        if size > 0:
            end_range = min(stat.size, size) - 1
            return self._client.read_range(self._internal_url, start=0, end=end_range)
        else:
            return self._client.read(self._internal_url)

    def read_range(
        self,
        start: int,
        end: int | None = None,
        check_exists: bool = False,
        headers: dict[str, str] | None = None,
    ) -> bytes:
        """Read the specified range of the resource and return the bytes read.

        Parameters
        ----------
        start : `int`
            Position of the first byte to read.
        end : `int`, optional
            Position of the last byte to read.
        check_exists : `bool`, optional
            Check the file exists before sending the GET request, which may
            fail if the file does not exist. This is useful when the caller
            has already checked the file exists before doing several
            partial reads, so we want to avoid checking for every call.
        headers : `dict[str, str]`, optional
            Headers to include in the partial GET request.
        """
        if check_exists:
            stat = self._stat()
            if not stat.is_file:
                raise FileNotFoundError(f"No file found at {self}")

            if stat.size == 0:
                # This is an empty file.
                return b""

        headers = {} if headers is None else dict(headers)
        headers.update({"Accept-Encoding": "identity"})
        return self._client.read_range(self._internal_url, start=start, end=end, headers=headers)

    def _as_local(self, multithreaded: bool = True, tmpdir: ResourcePath | None = None) -> tuple[str, bool]:
        """Download object and place in temporary directory.

        Parameters
        ----------
        multithreaded : `bool`, optional
            If `True` the transfer will be allowed to attempt to improve
            throughput by using parallel download streams. This may of no
            effect if the URI scheme does not support parallel streams or
            if a global override has been applied. If `False` parallel
            streams will be disabled.
        tmpdir : `ResourcePath` or `None`, optional
            Explicit override of the temporary directory to use for remote
            downloads.

        Returns
        -------
        path : `str`
            Path to local temporary file.
        temporary : `bool`
            Always returns `True`. This is always a temporary file.
        """
        # We need to ensure that this resource is actually a file. dCache
        # responds with a HTML-formatted content to a HTTP GET request to a
        # directory, which is not what we want.
        stat = self._stat()
        if not stat.is_file:
            raise FileNotFoundError(f"No file found at {self}")

        if tmpdir is None:
            local_dir, buffer_size = self._config.tmpdir_buffersize
            tmpdir = ResourcePath(local_dir, forceDirectory=True)
        else:
            buffer_size = _calc_tmpdir_buffer_size(tmpdir.ospath)

        with ResourcePath.temporary_uri(suffix=self.getExtension(), prefix=tmpdir, delete=False) as tmp_uri:
            self._client.download(self._internal_url, tmp_uri.ospath, buffer_size)
            return tmp_uri.ospath, True

    def write(self, data: BinaryIO | bytes, overwrite: bool = True) -> None:
        """Write the supplied bytes to the new resource.

        Parameters
        ----------
        data : `bytes`
            The bytes to write to the resource. The entire contents of the
            resource will be replaced.
        overwrite : `bool`, optional
            If `True` the resource will be overwritten if it exists. Otherwise
            the write will fail.
        """
        if self.isdir():
            raise ValueError(f"Method write() is not implemented for directory {self}")

        stat = self._stat()
        if stat.is_file and not overwrite:
            raise FileExistsError(f"File {self} exists and overwrite has been disabled")

        # Create parent directory and upload the data.
        self.parent().mkdir()
        self._client.write(self._internal_url, data)
        self._invalidate_metatada_cache()

    def remove(self) -> None:
        """Remove the resource.

        If the resource is a directory, it must be empty otherwise this
        method raises. Removing a non-existent file or directory is not
        considered an error.
        """
        if not self.exists():
            return

        if self.isdir():
            entries = self._client.read_dir(self._internal_url)
            if len(entries) > 0:
                raise IsADirectoryError(f"directory {self} is not empty")

        # This resource is a either file or an empty directory, we can remove
        # it.
        self._client.delete(self._internal_url)
        self._invalidate_metatada_cache()

    def remove_dir(self, recursive: bool = False) -> None:
        """Remove a directory if empty.

        Parameters
        ----------
        recursive: `bool`
            If `True` recursively remove all files and directories under this
            directory.

        Notes
        -----
            This method is not present in the superclass.
        """
        if not self.isdir():
            raise NotADirectoryError(f"{self} is not a directory")

        for root, subdirs, files in self.walk():
            if not recursive and (len(subdirs) > 0 or len(files) > 0):
                raise IsADirectoryError(f"directory {self} is not empty and recursive argument is False")

            for file in files:
                root.join(file).remove()

            for subdir in subdirs:
                DavResourcePath(root.join(subdir, forceDirectory=True)).remove_dir(recursive=recursive)

        # Remove empty top directory
        self.remove()

    def transfer_from(
        self,
        src: ResourcePath,
        transfer: str = "copy",
        overwrite: bool = False,
        transaction: TransactionProtocol | None = None,
        multithreaded: bool = True,
    ) -> None:
        """Transfer to this URI from another.

        Parameters
        ----------
        src : `ResourcePath`
            Source URI.
        transfer : `str`
            Mode to use for transferring the resource. Generically there are
            many standard options: copy, link, symlink, hardlink, relsymlink.
            Not all URIs support all modes.
        overwrite : `bool`, optional
            Allow an existing file to be overwritten. Defaults to `False`.
        transaction : `~lsst.resources.utils.TransactionProtocol`, optional
            A transaction object that can (depending on implementation)
            rollback transfers on error.  Not guaranteed to be implemented.
        multithreaded : `bool`, optional
            If `True` the transfer will be allowed to attempt to improve
            throughput by using parallel download streams. This may of no
            effect if the URI scheme does not support parallel streams or
            if a global override has been applied. If `False` parallel
            streams will be disabled.
        """
        # Fail early to prevent delays if remote resources are requested.
        if transfer not in self.transferModes:
            raise ValueError(f"Transfer mode {transfer} not supported by URI scheme {self.scheme}")

        # Existence checks cost time so do not call this unless we know
        # that debugging is enabled.
        destination_exists = self.exists()
        if log.isEnabledFor(logging.DEBUG):
            log.debug(
                "Transferring %s [exists: %s] -> %s [exists: %s] (transfer=%s)",
                src,
                src.exists(),
                self,
                destination_exists,
                transfer,
            )

        # Short circuit immediately if the URIs are identical.
        if self == src:
            log.debug(
                "Target and destination URIs are identical: %s, returning immediately."
                " No further action required.",
                self,
            )
            return

        if not overwrite and destination_exists:
            raise FileExistsError(f"Destination path {self} already exists.")

        if transfer == "auto":
            transfer = self.transferDefault

        # We can use webDAV 'COPY' or 'MOVE' if both the current and source
        # resources are located in the same server.
        if isinstance(src, type(self)) and self.root_uri() == src.root_uri():
            log.debug("Transfer from %s to %s directly", src, self)
            return (
                self._move_from(src, overwrite=overwrite)
                if transfer == "move"
                else self._copy_from(src, overwrite=overwrite)
            )

        # For resources of different classes we can perform the copy or move
        # operation by downloading to a local file and uploading to the
        # destination.
        self._copy_via_local(src)

        # This was an explicit move, try to remove the source.
        if transfer == "move":
            src.remove()

    def _copy_via_local(self, source: ResourcePath) -> None:
        """Replace the contents of this resource with the contents of a remote
        resource by using a local temporary file.

        Parameters
        ----------
        source : `ResourcePath`
            The source of the contents to copy to `self`.
        """
        with source.as_local() as local_uri:
            log.debug("Transfer from %s to %s via local file %s", source, self, local_uri)
            with open(local_uri.ospath, "rb") as f:
                self.write(data=f)

        self._invalidate_metatada_cache()

    def _copy_from(self, source: DavResourcePath, overwrite: bool = False) -> None:
        """Copy the contents of `source` to this resource. `source` must
        be a file.
        """
        # Copy is only supported for files, not directories.
        if source.isdir():
            raise ValueError(f"Copy is not supported for directory {source}")

        src_stat = source._stat()
        if not src_stat.is_file:
            raise FileNotFoundError(f"No such file {source}")

        dst_stat = self._stat()
        if dst_stat.is_dir:
            raise ValueError(f"Copy is not supported because destination {self} is a directory")

        self.parent().mkdir()
        self._client.copy(source._internal_url, self._internal_url, overwrite)
        self._invalidate_metatada_cache()

    def _move_from(self, source: DavResourcePath, overwrite: bool = False) -> None:
        """Send a MOVE webDAV request to replace the contents of this resource
        with the contents of another resource located in the same server.

        Parameters
        ----------
        source : `DavResourcePath`
            The source of the contents to move to `self`.
        """
        # Move is only supported for files, not directories.
        if source.isdir():
            raise ValueError(f"Move is not supported for directory {source}")

        src_stat = source._stat()
        if not src_stat.is_file:
            raise FileNotFoundError(f"No such file {source}")

        dst_stat = self._stat()
        if dst_stat.is_dir:
            raise ValueError(f"Move is not supported for destination directory {self}")

        # Create the destination's parent directory, move the source to
        # this resource and invalidate caches for both.
        self.parent().mkdir()
        self._client.move(source._internal_url, self._internal_url, overwrite)
        self._invalidate_metatada_cache()
        source._invalidate_metatada_cache()

    def walk(
        self, file_filter: str | re.Pattern | None = None
    ) -> Iterator[list | tuple[ResourcePath, list[str], list[str]]]:
        """Walk the directory tree returning matching files and directories.

        Parameters
        ----------
        file_filter : `str` or `re.Pattern`, optional
            Regex to filter out files from the list before it is returned.

        Yields
        ------
        dirpath : `ResourcePath`
            Current directory being examined.
        dirnames : `list` of `str`
            Names of subdirectories within dirpath.
        filenames : `list` of `str`
            Names of all the files within dirpath.
        """
        if not self.isdir():
            raise ValueError("Can not walk a non-directory URI")

        # We must return no entries for non-existent directories.
        if not self._stat().exists:
            return

        # Retrieve the entries in this directory
        entries = self._client.read_dir(self._internal_url)
        files = [e.name for e in entries if e.is_file]
        subdirs = [e.name for e in entries if e.is_dir]

        # Filter files
        if isinstance(file_filter, str):
            file_filter = re.compile(file_filter)

        if file_filter is not None:
            files = [f for f in files if file_filter.search(f)]

        if not subdirs and not files:
            return
        else:
            yield type(self)(self, forceAbsolute=False, forceDirectory=True), subdirs, files

        for subdir in subdirs:
            new_uri = self.join(subdir, forceDirectory=True)
            yield from new_uri.walk(file_filter)

    def generate_presigned_get_url(self, *, expiration_time_seconds: int) -> str:
        """Return a pre-signed URL that can be used to retrieve this resource
        using an HTTP GET without supplying any access credentials.

        Parameters
        ----------
        expiration_time_seconds : `int`
            Number of seconds until the generated URL is no longer valid.

        Returns
        -------
        url : `str`
            HTTP URL signed for GET.
        """
        return self._client.generate_presigned_get_url(self._internal_url, expiration_time_seconds)

    def generate_presigned_put_url(self, *, expiration_time_seconds: int) -> str:
        """Return a pre-signed URL that can be used to upload a file to this
        path using an HTTP PUT without supplying any access credentials.

        Parameters
        ----------
        expiration_time_seconds : `int`
            Number of seconds until the generated URL is no longer valid.

        Returns
        -------
        url : `str`
            HTTP URL signed for PUT.
        """
        return self._client.generate_presigned_put_url(self._internal_url, expiration_time_seconds)

    def to_fsspec(self) -> tuple[DavFileSystem, str]:
        """Return an abstract file system and path that can be used by fsspec.

        Returns
        -------
        fs : `fsspec.spec.AbstractFileSystem`
            A file system object suitable for use with the returned path.
        path : `str`
            A path that can be opened by the file system object.
        """
        if fsspec is None or not self._client._config.enable_fsspec:
            raise ImportError("fsspec is not available")

        path: str = self.path
        return DavFileSystem(self, path), path

    @contextlib.contextmanager
    def _openImpl(
        self,
        mode: str = "r",
        *,
        encoding: str | None = None,
    ) -> Iterator[ResourceHandleProtocol]:
        if self.isdir():
            raise OSError(f"open is not implemented for directory {self}")

        if mode in ("rb", "r") and self._client.accepts_ranges(self._internal_url):
            stat: DavFileMetadata = self._stat(refresh=True)
            if not stat.exists:
                raise FileNotFoundError(f"No such file {self}")

            if not stat.is_file:
                raise OSError(f"open is not implemented for directory {self}")

            handle: ResourceHandleProtocol = DavReadResourceHandle(mode, log, self, stat)
            if mode == "r":
                # cast because the protocol is compatible, but does not have
                # BytesIO in the inheritance tree
                yield io.TextIOWrapper(cast(Any, handle), encoding=encoding)
            else:
                yield handle
        else:
            with super()._openImpl(mode, encoding=encoding) as handle:
                yield handle


class DavFileSystem(AbstractFileSystem):
    """Minimal fsspec-compatible read-only file system which contains a single
    file.

    Parameters
    ----------
    uri : `DavResourcePath`
        URI of the single resource contained in the file system.

    path : `str`
        Path within the file system of the file.
    """

    def __init__(self, uri: DavResourcePath, path: str):
        self._uri: DavResourcePath = uri
        self._path: str = path

    def info(self, path: str, **kwargs: Any) -> dict[str, Any]:
        if path != self._path:
            raise FileNotFoundError(path)

        return {
            "name": path,
            "size": self._uri.size(),
            "type": "file",
        }

    def ls(self, path: str, detail: bool = True, **kwargs: Any) -> list[str] | list[dict[str, str]]:
        if path != self._path:
            raise FileNotFoundError(path)

        return list(self.info(path)) if detail else list(path)

    def modified(self, path: str) -> datetime.datetime:
        if path != self._path:
            raise FileNotFoundError(path)

        return self._uri._stat().last_modified

    def size(self, path: str) -> int:
        if path != self._path:
            raise FileNotFoundError(path)

        return self._uri.size()

    def isfile(self, path: str) -> bool:
        return path == self._path

    def isdir(self, path: str) -> bool:
        return False

    def exists(self, path: str, **kwargs: Any) -> bool:
        return path == self._path

    def open(
        self,
        path: str,
        mode: str = "rb",
        encoding: str | None = None,
        block_size: int | None = None,
        cache_options: dict[Any, Any] | None = None,
        compression: str | None = None,
        **kwargs: Any,
    ) -> ResourceHandleProtocol[Any]:
        if path != self._path:
            raise FileNotFoundError(path)

        with self._uri.open(mode=mode, encoding=encoding) as handle:
            return handle

    @property
    def fsid(self) -> Any:
        raise NotImplementedError

    def mkdir(self, path: str, create_parents: bool = True, **kwargs: Any) -> None:
        raise NotImplementedError

    def makedirs(self, path: str, exist_ok: bool = False) -> None:
        raise NotImplementedError

    def rmdir(self, path: str) -> None:
        raise NotImplementedError

    def walk(
        self,
        path: str,
        maxdepth: int | None = None,
        topdown: bool = True,
        on_error: str = "omit",
        **kwargs: Any,
    ) -> None:
        raise NotImplementedError

    def find(
        self,
        path: str,
        maxdepth: int | None = None,
        withdirs: bool = False,
        detail: bool = False,
        **kwargs: Any,
    ) -> None:
        raise NotImplementedError

    def du(
        self,
        path: str,
        total: bool = True,
        maxdepth: int | None = None,
        withdirs: bool = False,
        **kwargs: Any,
    ) -> None:
        raise NotImplementedError

    def glob(self, path: str, maxdepth: int | None = None, **kwargs: Any) -> None:
        raise NotImplementedError

    def rm_file(self, path: str) -> None:
        raise NotImplementedError

    def rm(self, path: str, recursive: bool = False, maxdepth: int | None = None) -> None:
        raise NotImplementedError

    def touch(self, path: str, truncate: bool = True, **kwargs: Any) -> None:
        raise NotImplementedError

    def ukey(self, path: str) -> None:
        raise NotImplementedError

    def created(self, path: str) -> None:
        raise NotImplementedError



# This file is part of lsst-resources.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# Use of this source code is governed by a 3-clause BSD-style
# license that can be found in the LICENSE file.

from __future__ import annotations

__all__ = ("NoTransaction", "TransactionProtocol", "get_tempdir", "os2posix", "posix2os")

import contextlib
import logging
import multiprocessing
import os
import posixpath
import shutil
import stat
import tempfile
from collections.abc import Callable, Iterator
from functools import cache
from pathlib import Path, PurePath, PurePosixPath
from typing import Any, Protocol

# Determine if the path separator for the OS looks like POSIX
IS_POSIX = os.sep == posixpath.sep

# Root path for this operating system. This can use getcwd which
# can fail in some situations so in the default case assume that
# posix means posix and only determine explicitly in the non-posix case.
OS_ROOT_PATH = posixpath.sep if IS_POSIX else Path().resolve().root

# Maximum number of worker threads for parallelized operations.
# If greater than 10, be aware that this number has to be consistent
# with connection pool sizing (for example in urllib3).
MAX_WORKERS = 10

log = logging.getLogger(__name__)


def os2posix(ospath: str) -> str:
    """Convert a local path description to a POSIX path description.

    Parameters
    ----------
    ospath : `str`
        Path using the local path separator.

    Returns
    -------
    posix : `str`
        Path using POSIX path separator.
    """
    if IS_POSIX:
        return ospath

    posix = PurePath(ospath).as_posix()

    # PurePath strips trailing "/" from paths such that you can no
    # longer tell if a path is meant to be referring to a directory
    # Try to fix this.
    if ospath.endswith(os.sep) and not posix.endswith(posixpath.sep):
        posix += posixpath.sep

    return posix


def posix2os(posix: PurePath | str) -> str:
    """Convert a POSIX path description to a local path description.

    Parameters
    ----------
    posix : `str`, `~pathlib.PurePath`
        Path using the POSIX path separator.

    Returns
    -------
    ospath : `str`
        Path using OS path separator.
    """
    if IS_POSIX:
        return str(posix)

    posixPath = PurePosixPath(posix)
    paths = list(posixPath.parts)

    # Have to convert the root directory after splitting
    if paths[0] == posixPath.root:
        paths[0] = OS_ROOT_PATH

    # Trailing "/" is stripped so we need to add back an empty path
    # for consistency
    if str(posix).endswith(posixpath.sep):
        paths.append("")

    return os.path.join(*paths)


@cache
def get_tempdir() -> str:
    """Get POSIX path to temporary directory.

    Returns
    -------
    tmpdir : `str`
        Path to the default temporary directory location.

    Notes
    -----
    Uses the value of environment variables ``LSST_RESOURCES_TMPDIR`` or
    ``TMPDIR``, if defined. Otherwise use the system temporary directory,
    with a last-resort fallback to the current working directory if
    nothing else is available.
    """
    tmpdir = None
    # $TMPDIR is also checked with getttempdir() below.
    for dir in (os.getenv(v) for v in ("LSST_RESOURCES_TMPDIR", "TMPDIR")):
        if dir and os.path.isdir(dir):
            tmpdir = dir
            break

    if tmpdir is None:
        tmpdir = tempfile.gettempdir()

    return tmpdir


class NoTransaction:
    """A simple emulation of the
    `~lsst.daf.butler.core.datastore.DatastoreTransaction` class.

    Notes
    -----
    Does nothing. Used as a fallback in the absence of an explicit transaction
    class.
    """

    def __init__(self) -> None:
        return

    @contextlib.contextmanager
    def undoWith(self, name: str, undoFunc: Callable, *args: Any, **kwargs: Any) -> Iterator[None]:
        """No-op context manager to replace
        `~lsst.daf.butler.core.datastore.DatastoreTransaction`.

        Parameters
        ----------
        name : `str`
            The name of this undo request.
        undoFunc : `~collections.abc.Callable`
            Function to call if there is an exception. Not used.
        *args : `~typing.Any`
            Parameters to pass to ``undoFunc``.
        **kwargs : `~typing.Any`
            Keyword parameters to pass to ``undoFunc``.

        Yields
        ------
        `None`
            Context manager returns nothing since transactions are disabled
            by definition.
        """
        yield None


class TransactionProtocol(Protocol):
    """Protocol for type checking transaction interface."""

    @contextlib.contextmanager
    def undoWith(self, name: str, undoFunc: Callable, *args: Any, **kwargs: Any) -> Iterator[None]: ...


def makeTestTempDir(default_base: str | None = None) -> str:
    """Create a temporary directory for test usage.

    The directory will be created within ``LSST_RESOURCES_TEST_TMP`` if that
    environment variable is set, falling back to ``LSST_RESOURCES_TMPDIR``
    amd then ``default_base`` if none are set.

    Parameters
    ----------
    default_base : `str`, optional
        Default parent directory. Will use system default if no environment
        variables are set and base is set to `None`.

    Returns
    -------
    dir : `str`
        Name of the new temporary directory.
    """
    base = default_base
    for envvar in ("LSST_RESOURCES_TEST_TMP", "LSST_RESOURCES_TMPDIR"):
        if envvar in os.environ and os.environ[envvar]:
            base = os.environ[envvar]
            break
    return tempfile.mkdtemp(dir=base)


def removeTestTempDir(root: str | None) -> None:
    """Attempt to remove a temporary test directory, but do not raise if
    unable to.

    Unlike `tempfile.TemporaryDirectory`, this passes ``ignore_errors=True``
    to ``shutil.rmtree`` at close, making it safe to use on NFS.

    Parameters
    ----------
    root : `str`, optional
        Name of the directory to be removed.  If `None`, nothing will be done.
    """
    if root is not None and os.path.exists(root):
        shutil.rmtree(root, ignore_errors=True)


def ensure_directory_is_writeable(directory_path: str | bytes) -> None:
    """Given the path to a directory, ensures that we are able to write it and
    access files in it.

    Alters the directory permissions by adding the owner-write and
    owner-traverse permission bits if they aren't already set

    Parameters
    ----------
    directory_path : `str` or `bytes`
        Path to the directory that will be made writeable.
    """
    current_mode = os.stat(directory_path).st_mode
    desired_mode = current_mode | stat.S_IWUSR | stat.S_IXUSR
    if current_mode != desired_mode:
        os.chmod(directory_path, desired_mode)


def _get_int_env_var(env_var: str) -> int | None:
    int_value = None
    env_value = os.getenv(env_var)
    if env_value is not None:
        with contextlib.suppress(TypeError):
            int_value = int(env_value)
    return int_value


@cache
def _get_num_workers() -> int:
    f"""Calculate the number of workers to use.

    Returns
    -------
    num : `int`
        The number of workers to use. Will use the value of the
        ``LSST_RESOURCES_NUM_WORKERS`` environment variable if set. Will fall
        back to using the CPU count (plus 2) but capped at {MAX_WORKERS}.
    """
    num_workers: int | None = None
    num_workers = _get_int_env_var("LSST_RESOURCES_NUM_WORKERS")

    # If someone is explicitly specifying a number, let them use that number.
    if num_workers is not None:
        return num_workers

    if num_workers is None:
        # CPU_LIMIT is used on nublado.
        cpu_limit = _get_int_env_var("CPU_LIMIT") or multiprocessing.cpu_count()
        if cpu_limit is not None:
            num_workers = cpu_limit + 2

    # But don't ever return more than the maximum allowed.
    return min([num_workers, MAX_WORKERS])



# This file is part of lsst-resources.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# Use of this source code is governed by a 3-clause BSD-style
# license that can be found in the LICENSE file.
from __future__ import annotations

__all__ = ("FileResourceHandle",)

import logging
from collections.abc import Iterable
from io import SEEK_SET
from typing import IO, TYPE_CHECKING, AnyStr, TypeVar

from ._baseResourceHandle import BaseResourceHandle

if TYPE_CHECKING:
    from .._resourcePath import ResourcePath

U = TypeVar("U", str, bytes)


class FileResourceHandle(BaseResourceHandle[U]):
    """File based specialization of `.BaseResourceHandle`.

    Parameters
    ----------
    mode : `str`
        Handle modes as described in the python `io` module.
    log : `~logging.Logger`
        Logger to used when writing messages.
    uri : `lsst.resources.ResourcePath`
        URI of the file on the filesystem to use.
    encoding : `str` or None
        Optionally supply the encoding of the file.
    newline : `str`
        When doing multiline operations, break the stream on given character.
        Defaults to newline. If a file is opened in binary mode, this argument
        is not used, as binary files will only split lines on the binary
        newline representation.

    Notes
    -----
    Documentation on the methods of this class line should refer to the
    corresponding methods in the `io` module.
    """

    def __init__(
        self, mode: str, log: logging.Logger, uri: ResourcePath, *, encoding: str | None, newline: str = "\n"
    ):
        super().__init__(mode, log, uri, newline=newline)
        self._filename = uri.ospath
        # opening a file in binary mode does not support a newline argument
        newline_arg = None if "b" in mode else newline
        self._fileHandle: IO = open(file=uri.ospath, mode=self._mode, newline=newline_arg, encoding=encoding)

    @property
    def name(self) -> str:
        # More consistent to return the path without the file://.
        return self._uri.ospath

    @property
    def mode(self) -> str:
        return self._mode

    def close(self) -> None:
        self._fileHandle.close()

    @property
    def closed(self) -> bool:
        return self._fileHandle.closed

    def fileno(self) -> int:
        return self._fileHandle.fileno()

    def flush(self) -> None:
        self._fileHandle.flush()

    @property
    def isatty(self) -> bool:
        return self._fileHandle.isatty()

    def readable(self) -> bool:
        return self._fileHandle.readable()

    def readline(self, size: int = -1) -> U:
        return self._fileHandle.readline(size)
        ...

    def readlines(self, hint: int = -1) -> Iterable[U]:
        return self._fileHandle.readlines(hint)

    def seek(self, offset: int, whence: int = SEEK_SET) -> int:
        return self._fileHandle.seek(offset, whence)

    def seekable(self) -> bool:
        return self._fileHandle.seekable()

    def tell(self) -> int:
        return self._fileHandle.tell()

    def truncate(self, size: int | None = None) -> int:
        return self._fileHandle.truncate(size)

    def writable(self) -> bool:
        return self._fileHandle.writable()

    def writelines(self, lines: Iterable[AnyStr]) -> None:
        self._fileHandle.writelines(lines)

    def read(self, size: int = -1) -> U:
        return self._fileHandle.read(size)

    def write(self, b: U) -> int:
        return self._fileHandle.write(b)
